#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for casa_envoy.palaver
"""
import argparse
import itertools
from unittest.mock import MagicMock, patch

import casa_envoy.palaver as palaver
import pytest
from casa_envoy.launchers import CalibrationLauncher, ImagingLauncher
from pycapo import CapoConfig

expected_settings = {
    "useCasa": False,
    "homeForReprocessing": "/home/casa/packages/pipeline/current",
    "rootDirectory": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "processingDirectory": "tmpo1ca1pp_",
    "parent_path": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_",
    "metadata": "test/input_files/test.json",
    "ppr": "test/input_files/PPR.xml",
    "product_type": "standard-cal",
}


@pytest.fixture
def args() -> argparse.Namespace:
    return palaver.arg_parser().parse_args([])


class TestPalaver:
    def test_get_settings(self, args):
        """
        verifies correct determination of needed settings

        :return:
        """
        args.standard_cal = ["test/input_files/test.json", "test/input_files/PPR.xml"]

        with patch(
            "casa_envoy.palaver.CapoConfig",
            MagicMock(return_value=CapoConfig(profile="test-capo", path="test/input_files/")),
        ):
            with patch(
                "pathlib.Path.cwd",
                MagicMock(return_value="/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_"),
            ) as cwd:
                settings = palaver._get_settings(cwd, ["test/input_files/test.json", "test/input_files/PPR.xml"])
                assert settings["useCasa"] == expected_settings["useCasa"]
                assert settings["homeForReprocessing"] == expected_settings["homeForReprocessing"]
                assert settings["metadata"] == expected_settings["metadata"]
                assert settings["ppr"] == expected_settings["ppr"]

    @patch("json.loads")
    def test_setup_launcher(self, mock_json, args):
        """
        Verifies that the correct type of launcher is created

        :return:
        """
        with patch(
            "pathlib.Path.cwd",
            MagicMock(return_value="/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_"),
        ) as cwd:
            args.standard_cal = ["test/input_files/test.json", "test/input_files/PPR.xml"]
            launcher = palaver._setup_launcher(cwd, args)
            assert isinstance(launcher, CalibrationLauncher)
            args.standard_cal = None

            args.standard_img = [
                "test/input_files/image-metadata.json",
                "test/input_files/cmsimage-PPR.xml",
            ]
            launcher = palaver._setup_launcher(cwd, args)
            assert isinstance(launcher, ImagingLauncher)

    @patch("json.loads")
    def test_setup_launcher_multiple_code_paths_specified(self, mock_json):
        """
        Verifies that multiple code-paths can't all be specified at once

        :return:
        """
        std_cal_args = ["--standard-cal", "test/input_files/test.json", "test/input_files/PPR.xml"]
        vlass_cal_args = ["--vlass-cal", "file", "other-file"]
        restore_args = ["--restore", "test/input_files/restore.json"]
        std_img_args = [
            "--standard-img",
            "test/input_files/cmsimage-PPR.xml",
            "test/input_files/image-metadata.json",
        ]
        vlass_img_args = ["--vlass-img", "file", "other-file"]

        with patch(
            "pathlib.Path.cwd",
            MagicMock(return_value="/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_"),
        ) as cwd:
            for one_codepath_args, two_codepath_args in itertools.combinations(
                [std_cal_args, vlass_cal_args, restore_args, std_img_args, vlass_img_args], 2
            ):
                raw_args = one_codepath_args + two_codepath_args
                parser = palaver.arg_parser()
                real_args = parser.parse_args(raw_args)
                with pytest.raises(ValueError, match="Expected exactly one"):
                    palaver._setup_launcher(cwd, real_args)

    def test_parser(self):
        """
        Verifies parser setup

        :return:
        """
        parser = palaver.arg_parser()
        assert isinstance(parser, argparse.ArgumentParser)

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_cal(self, mock_cwd, mock_chdir, mock_json, args):
        """
        verifies use of correct launcher for calibrations

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args.standard_cal = ["test/input_files/test.json", "test/input_files/PPR.xml"]

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.CalibrationLauncher.launch_casa") as cal_launcher:
                palaver.main()
                assert cal_launcher.call_count == 1

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_img(self, mock_cwd, mock_chdir, mock_json, args):
        """
        verifies use of correct launcher for images

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args.standard_img = [
            "test/input_files/image-metadata.json",
            "test/input_files/cmsimage-PPR.xml",
        ]

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.ImagingLauncher.launch_casa") as img_launcher:
                palaver.main()
                assert img_launcher.call_count == 1

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_restore(self, mock_cwd, mock_chdir, mock_json, args):
        """
        verifies use of correct launcher for restores

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args.restore = [
            "test/input_files/restore.json",
        ]

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.CalibrationLauncher.launch_casa") as cal_launcher:
                palaver.main()
                assert cal_launcher.call_count == 1
